module Parser(parser) where

import Char
import Parsek
import While

parser xs = case parse program longestResultWithLeftover xs of
  Left (x,_,_) -> error
      ("Parse error at "++case x of Nothing -> "eof"; Just y -> [y])
  Right (a,xs) -> case xs of
      [] -> a
      _  -> error ("Could not parse the whole input. Remaining input: "++xs)

ch c = do spaces; char c
keyword x = do spaces; string x
identifier = do spaces; many1 (satisfy isAlpha)

number = do spaces
            digits <- many1 (satisfy isDigit)
            return (foldl (\a d -> 10*a + digitToInt d) (0::Int) digits)

program = statement `chainl1` semicolonop

statement = do keyword "skip"
               return Skip
            <|>
            do idname <- identifier
               ch '='
               e <- aexp
               return (Assignment idname e)
            <|>
            do keyword "if"
               ch '('
               e <- bexp
               ch ')'
               stmt1 <- statement
               keyword "else"
               stmt2 <- statement
               return (If e stmt1 stmt2)
            <|>
            do keyword "while"
               e <- bexp
               stmt <- statement
               return (While e stmt)
            <|>
            do ch '('
               p <- program
               ch ')'
               return p

bexp = lfactor `chainl1` andop 
lfactor = do a1 <- aexp
             op <- relop
             a2 <- aexp
             return (op a1 a2)
         <|> 
         do keyword "not"
            f <- lfactor
            return (Not f)
         <|>
         do ch '('
            e <- bexp
            ch ')'
            return e
            
aexp = term `chainl1` addop 
term = factor `chainl1` mulop 
factor = do num <- number
            return (Num num)
         <|>
         do name <- identifier
            return (Var name)
         <|>
         do ch '('
            e <- aexp
            ch ')'
            return e
             
mulop = do ch '*'
           return Mul 
semicolonop = do ch ';'
                 return Sequence
andop = do ch '&'
           return And 

relop = do ch '='
           return Equal 
        <|>
        do keyword "<="
           return LessEqual 

addop = do ch '+'
           return Add 
        <|>
        do ch '-'
           return Sub
{- 
test1 = parser "skip"
test2 = parser "x=0"
test3 = parser "skip;skip"
test4 = parser "skip;skip;skip"
test5 = parser "skip;(skip;skip)"
test6 = parser "if(x=1)skip else skip"
test9 = parser "while(1<=n) (f=f*n; n=n-1)"
testall = [test1, test2, test3, test4, test5, test6, test9]
t = testall ==[Skip,Assignment "x" (Num 0),Sequence Skip Skip,
    Sequence (Sequence Skip Skip) Skip,Sequence Skip (Sequence Skip Skip),
    If (Equal (Var "x") (Num 1)) Skip Skip,
    While (LessEqual (Num 1) (Var "n")) (Sequence
    (Assignment "f" (Mul (Var "f") (Var "n")))
    (Assignment "n" (Sub (Var "n") (Num 1))))]
-}
-- Local Variables:
-- mode: hs
-- End:
