module While where

-- Type synonymes are introduced for readability
type Program = Stm
type IdName = String

data Stm = Skip
         | Assignment IdName Aexp
         | If Bexp Stm Stm
         | While Bexp Stm
	 | Sequence Stm Stm
         deriving (Show, Eq)

data Aexp = Num Int
          | Var IdName
          | Add Aexp Aexp
	  | Sub Aexp Aexp
	  | Mul Aexp Aexp
          deriving (Show, Eq)

data Bexp = Logical Bool
	  | And Bexp Bexp
	  | Not Bexp
	  | Equal Aexp Aexp
	  | LessEqual Aexp Aexp
          deriving (Show, Eq)

-- State and Substitution are more abstract but toString cannot be implemented
type State = IdName -> Int
type State' = [(IdName, Int)]

type Substitution a = IdName -> a
type Substitution' a = [(IdName, a)]

-- Use classes or different function names as in the value functions
class ToString a where
  toString :: a -> String

instance ToString Aexp where
  toString (Num n) = show n
  toString (Var name) = name
  toString (Add a1 a2) = "(" ++ toString a1 ++ "+" ++ toString a2 ++ ")"
  toString (Sub a1 a2) = "(" ++ toString a1 ++ "-" ++ toString a2 ++ ")"
  toString (Mul a1 a2) = "(" ++ toString a1 ++ "*" ++ toString a2 ++ ")"

instance ToString Bexp where
  toString (Logical b) = show b
  toString (Not b) = "!" ++ toString b
  toString (And b1 b2) = "(" ++ toString b1 ++ "&" ++ toString b2 ++ ")"
  toString (Equal a1 a2) = "(" ++ toString a1 ++ "=" ++ toString a2 ++ ")"
  toString (LessEqual a1 a2) = "(" ++ toString a1 ++ "<=" ++ toString a2 ++ ")"

instance ToString Stm where
  toString Skip = "skip"
  toString (Assignment name aexp) = name ++ "=" ++ toString aexp
  toString (Sequence stm1 stm2) =
         "(" ++ toString stm1 ++ "; " ++ toString stm2 ++ ")"
  toString (If bexp stm1 stm2) = "if " ++ toString bexp ++
         " then " ++ toString stm1  ++ " else " ++ toString stm2
  toString (While bexp stm) = "while " ++ toString bexp ++
         " do " ++ toString stm

aValue :: Aexp -> State -> Int

aValue (Num n) _ = n
aValue (Var name) state = state name
aValue (Add a1 a2) state = (aValue a1 state) + (aValue a2 state)
aValue (Sub a1 a2) state = (aValue a1 state) - (aValue a2 state)
aValue (Mul a1 a2) state = (aValue a1 state) * (aValue a2 state)

bValue :: Bexp -> State -> Bool

bValue (Logical b) _ = b
bValue (Not b) state = not (bValue b state)
bValue (And b1 b2) state = (bValue b1 state) && (bValue b2 state)
bValue (Equal a1 a2) state = (aValue a1 state) == (aValue a2 state)

class Free a where
  free :: a -> [IdName]

-- sets are represented as lists. If `union`  from the List library is 
-- used instead of ++ no duplicates will be introduced 
instance Free Aexp where
  free (Num n) = []
  free (Var name) = [name]
  free (Add a1 a2) = (free a1) ++ (free a2)
  free (Sub a1 a2) = (free a1) ++ (free a2)
  free (Mul a1 a2) = (free a1) ++ (free a2)
  
instance Free Bexp where
  free (Logical _) = []
  free (Equal a1 a2) = (free a1) ++ (free a2)
  free (LessEqual a1 a2) = (free a1) ++ (free a2)
  free (And b1 b2) = (free b1) ++ (free b2)
  free (Not b) = free b

class Subst a where
  subst :: Substitution Aexp -> a -> a

instance Subst Aexp where
  subst sbst (Num n) = Num n
  subst sbst (Var name) = sbst name
  subst sbst (Add a1 a2) = Add (subst sbst a1) (subst sbst a2)
  subst sbst (Sub a1 a2) = Sub (subst sbst a1) (subst sbst a2)
  subst sbst (Mul a1 a2) = Mul (subst sbst a1) (subst sbst a2)

instance Subst Bexp where
  subst sbst (Logical b) = Logical b
  subst sbst (Equal a1 a2) = Equal (subst sbst a1) (subst sbst a2)
  subst sbst (LessEqual a1 a2) = LessEqual (subst sbst a1) (subst sbst a2)
  subst sbst (And b1 b2) = And (subst sbst b1) (subst sbst b2)
  subst sbst (Not b) = Not (subst sbst b)

instance ToString State' where
  toString [] = ""
  toString ((name, value):ms) = "(" ++ name ++ ", " ++ (show value) ++ ")" ++
           toString ms

instance ToString (Substitution' Aexp) where
  toString [] = ""
  toString ((name, aexp):ms) = "(" ++ name ++ ", " ++ (toString aexp) ++ ")" ++
           toString ms

{-  this comment contains tests
test1 = toString ( While (LessEqual (Num 1) (Var "n")) (Sequence
    (Assignment "f" (Mul (Var "f") (Var "n")))
    (Assignment "n" (Sub (Var "n") (Num 1)))))

state1 "x" = 3
state1 "y" = 2
state1 _ = 0
exp1 = Equal (Add (Num 1) (Var "x")) (Mul (Var "y") (Num 2) )
test2 =  bValue exp1 state1

subst1 "x" = Num 3
subst1 "y" = Num 2

empty :: State
empty _ = 0

test3 = bValue (subst subst1 exp1) empty
-}
-- Local Variables:
-- mode: hs
-- End:

